"""Module that contains plugin's configurator."""

from __future__ import annotations

import os
import dataclasses

import tomlkit


CONFIG_SOURCES: dict[str, str] = {
    "pyproject.toml": "tool.poetry.plugins.dotenv",
    "os": "POETRY_PLUGIN_DOTENV_",
}

_STR_BOOLEAN_MAPPING: dict[str, bool] = {
    "y": True,
    "yes": True,
    "t": True,
    "on": True,
    "1": True,
    "true": True,
    "n": False,
    "no": False,
    "f": False,
    "off": False,
    "0": False,
    "false": False,
}


@dataclasses.dataclass
class _Config:
    """Defines the schema and default values for the plugin configuration."""

    ignore: bool = False
    location: str = ""


# TODO(pivoshenko): this configuration loader is a "quick patch" solution and needs refinement
class Config(_Config):
    """Configuration loader for the plugin."""

    def __init__(self, working_dir: str) -> None:
        """Initialize and load configuration from the defined sources."""

        super().__init__()

        source_config = {}
        for config_source, section in CONFIG_SOURCES.items():
            if config_source.endswith(".toml"):
                config = _load_config_from_toml(os.path.join(working_dir, config_source), section)

            elif config_source.endswith("os"):
                config = _load_config_from_os(section)

            else:  # pragma: no cover
                pass

            source_config.update(config)

        self._apply_source_config(source_config)

    def _apply_source_config(self, source_config: dict[str, str | bool | None]) -> None:
        """Apply the loaded configuration to the instance variables."""

        for field in self.__dataclass_fields__.values():
            source_value = source_config.get(field.name)

            if (
                isinstance(field.default, bool)
                and source_value
                and not isinstance(source_value, bool)
            ):
                source_value = _as_bool(source_value)

            if source_value is not None:
                setattr(self, field.name, source_value)


def _load_config_from_toml(filepath: str, section: str) -> dict[str, str | bool | None]:
    """Load configuration from a TOML file."""

    if not os.path.exists(filepath):
        return {}

    with open(filepath, "rb") as toml_file:
        config = tomlkit.load(toml_file)

    for key in section.split("."):
        config = config.get(key, {})

    return config if isinstance(config, dict) else {}


def _load_config_from_os(section: str) -> dict[str, str | bool | None]:
    """Load configuration from OS environment variables."""

    return {
        key[len(section) :].lower(): value
        for key, value in os.environ.items()
        if key.startswith(section)
    }


def _as_bool(value: str) -> bool:
    """Convert a string value to its Boolean equivalent.

    This function is inspired by ``distutils strtobool``.
    """

    try:
        return _STR_BOOLEAN_MAPPING[value.lower()]

    except KeyError:
        msg = f"Invalid truth value '{value}'"
        raise ValueError(msg) from None



"""Module that contains custom exceptions that are used across plugin."""

from __future__ import annotations


class PoetryPluginDotenvError(Exception):
    """Base exception."""


class PoetryPluginDotenvPatternError(PoetryPluginDotenvError):
    """Exception raised when a dotenv pattern is not valid or not found."""



"""Package for the dotenv functionality.

This package is a fork of the ``python-dotenv`` package.

References
----------
#. `python-dotenv <https://github.com/theskumar/python-dotenv>`_

"""

from __future__ import annotations

from poetry_plugin_dotenv.dotenv import core
from poetry_plugin_dotenv.dotenv import parsers
from poetry_plugin_dotenv.dotenv import variables


__all__ = ["core", "parsers", "variables"]



"""Module that contains plugin's logger."""

from __future__ import annotations

import enum

from functools import partial
from typing import TYPE_CHECKING

from cleo.io.outputs.output import Verbosity


if TYPE_CHECKING:  # pragma: no cover
    from cleo.events.console_command_event import ConsoleCommandEvent


class Style(enum.Enum):
    """Style tags to color text output."""

    INFO = "<info>{0!s}</info>"
    DEBUG = "<debug>{0!s}</debug>"
    WARNING = "<warning>{0!s}</warning>"
    ERROR = "<error>{0!s}</error>"

    def __call__(self, text: str) -> str:
        """Apply the style to the text."""

        return self.value.format(text)


class Logger:
    """Logger for the ``poetry`` events.

    Notes
    -----
    Since this logger is used by a plugin that runs before the main command,
    all messages are logged only in debug mode.

    """

    def __init__(self, event: ConsoleCommandEvent) -> None:
        """Initialize the logger."""

        self.write_line = partial(event.io.write_line, verbosity=Verbosity.DEBUG)

    def info(self, msg: str) -> None:
        """Log an informational message."""

        self._log(Style.INFO, msg)

    def debug(self, msg: str) -> None:
        """Log a debug message."""

        self._log(Style.DEBUG, msg)  # pragma: nocover

    def warning(self, msg: str) -> None:
        """Log a warning message."""

        self._log(Style.WARNING, msg)

    def error(self, msg: str) -> None:
        """Log an error message."""

        self._log(Style.ERROR, msg)

    def _log(self, style: Style, msg: str) -> None:
        """Log a message with the given style."""

        self.write_line(style(msg))



"""poetry-plugin-dotenv - is the plugin that automatically loads environment variables from a dotenv file into the environment before poetry commands are run."""

from __future__ import annotations


__title__ = "poetry-plugin-dotenv"
__summary__ = "poetry-plugin-dotenv - is the plugin that automatically loads environment variables from a dotenv file into the environment before poetry commands are run."
__uri__ = "https://github.com/pivoshenko/poetry-plugin-dotenv"

__version__ = "2.7.1"

__author__ = "Volodymyr Pivoshenko"
__email__ = "volodymyr.pivoshenko@gmail.com"

__license__ = "MIT"
__copyright__ = "Copyright 2025, Volodymyr Pivoshenko"

from poetry_plugin_dotenv import config
from poetry_plugin_dotenv import dotenv
from poetry_plugin_dotenv import plugin


__all__ = ["config", "dotenv", "plugin"]



"""Module that contains core dotenv functionality."""

from __future__ import annotations

import io
import os
import sys
import contextlib

from collections import OrderedDict
from typing import IO
from typing import TYPE_CHECKING

from poetry_plugin_dotenv.dotenv import parsers
from poetry_plugin_dotenv.dotenv import variables


if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterable
    from collections.abc import Iterator


class DotEnv:
    """Model of a dotenv file."""

    def __init__(
        self,
        filepath: str | None = None,
        stream: IO[str] | None = None,
        *,
        interpolate: bool = True,
        override: bool = True,
    ) -> None:
        """Initialize."""

        self.filepath = filepath
        self.stream = stream
        self.override = override
        self.interpolate = interpolate
        self.encoding = "utf-8"
        self._dict: OrderedDict[str, str] | None = None

    def dict(self) -> OrderedDict[str, str]:
        """Get content of a dotenv file."""

        if self._dict:
            return self._dict

        raw_values = self.parse()

        if self.interpolate:
            self._dict = resolve(raw_values, override=self.override)

        else:
            self._dict = OrderedDict(raw_values)

        return self._dict

    def parse(self) -> Iterator[tuple[str, str]]:
        """Parse a dotenv file."""

        with self._get_stream() as stream:
            for mapping in parsers.parse_stream(stream):
                if mapping.key is not None:
                    yield mapping.key, mapping.value  # type: ignore[misc]

    def set_as_environment_variables(self) -> bool:
        """Load current dotenv as system environment variables."""

        if self.dict():
            for key, value in self.dict().items():
                if key in os.environ and not self.override:
                    continue

                if value:
                    os.environ[key] = value

            return True

        return False  # pragma: nocover

    @contextlib.contextmanager
    def _get_stream(self) -> Iterator[IO[str]]:
        """Get a dotenv stream."""

        if self.filepath and os.path.isfile(self.filepath):
            with open(self.filepath, encoding=self.encoding) as stream:
                yield stream

        elif self.stream:
            yield self.stream

        else:
            yield io.StringIO("")  # pragma: nocover


def resolve(
    values: Iterable[tuple[str, str]],
    *,
    override: bool,
) -> OrderedDict[str, str]:
    """Resolve dotenv variables."""

    new_values: OrderedDict[str, str] = OrderedDict()

    for name, value in values:
        if value is None:
            result = None  # pragma: nocover

        else:
            atoms = variables.parse(value)
            env: OrderedDict[str, str] = OrderedDict()

            if override:
                env.update(os.environ)
                env.update(new_values)

            else:
                env.update(new_values)
                env.update(os.environ)

            result = "".join(atom.resolve(env) for atom in atoms)

        new_values[name] = result

    return new_values


def walk_to_root(path: str) -> Iterator[str]:
    """Yield directories starting from the given directory up to the root."""

    if not os.path.exists(path):
        msg = "Starting path not found."
        raise OSError(msg)

    current_dir = os.path.abspath(os.path.dirname(path) if os.path.isfile(path) else path)

    while True:
        yield current_dir
        parent_dir = os.path.abspath(os.path.join(current_dir, os.pardir))

        if current_dir == parent_dir:
            break

        current_dir = parent_dir


def find(filename: str = ".env", *, usecwd: bool = False) -> str:
    """Search in increasingly higher folders for the given file."""

    path = (
        os.getcwd()
        if usecwd or getattr(sys, "frozen", False)
        else os.path.dirname(os.path.abspath(sys._getframe(1).f_code.co_filename))  # noqa: SLF001
    )

    for dirname in walk_to_root(path):
        check_path = os.path.join(dirname, filename)

        if os.path.isfile(check_path):
            return check_path

    return ""


def load(
    filepath: str | None = None,
    stream: IO[str] | None = None,
    *,
    interpolate: bool = True,
    override: bool = True,
) -> bool:
    """Parse a dotenv file and then load all the variables found as environment variables."""

    dotenv = DotEnv(filepath=filepath, interpolate=interpolate, override=override, stream=stream)
    return dotenv.set_as_environment_variables()


def values(
    filepath: str | None = None,
    stream: IO[str] | None = None,
    *,
    interpolate: bool = True,
) -> OrderedDict[str, str]:
    """Parse a dotenv file and return its content as a dictionary."""

    return DotEnv(filepath=filepath, stream=stream, interpolate=interpolate).dict()



"""Module that contains models of the dotenv variables and literals."""

from __future__ import annotations

import re
import dataclasses

from typing import TYPE_CHECKING


if TYPE_CHECKING:  # pragma: no cover
    from collections import OrderedDict
    from collections.abc import Iterator


_posix_variable = re.compile(
    r"""
    \$\{
        (?P<name>[^}:]*)
        (?::-
            (?P<default>[^}]*)
        )?
    }
    """,
    re.VERBOSE,
)


@dataclasses.dataclass(frozen=True)
class Literal:
    """Represents a literal value."""

    value: str

    def resolve(self, *args, **kwargs) -> str:  # type: ignore[no-untyped-def]
        """Get a literal value."""

        return self.value


@dataclasses.dataclass(frozen=True)
class Variable:
    """Represents a variable with an optional default value."""

    name: str
    default: str | None = None

    def resolve(self, env: OrderedDict[str, str], *args, **kwargs) -> str:  # type: ignore[no-untyped-def]
        """Resolve the variable value from the environment or use the default."""

        return env.get(self.name, self.default) or ""  # type: ignore[arg-type]


def parse(value: str) -> Iterator[Literal | Variable]:
    """Parse a string and yield literals and variables."""

    cursor = 0
    value_length = len(value)
    matches = list(_posix_variable.finditer(value))

    for match in matches:
        start, end = match.span()
        name = match.group("name")
        default = match.group("default")

        if start > cursor:
            yield Literal(value=value[cursor:start])

        yield Variable(name=name, default=default)
        cursor = end

    if cursor < value_length:
        yield Literal(value=value[cursor:])



"""Module that contains dotenv parsers."""

from __future__ import annotations

import re
import codecs
import dataclasses

from typing import IO
from typing import TYPE_CHECKING

from poetry_plugin_dotenv.exceptions import PoetryPluginDotenvPatternError


if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterator


def make_regex(string: str, extra_flags: int = 0) -> re.Pattern[str]:
    """Construct a regex expression."""

    return re.compile(string, re.UNICODE | extra_flags)


_newline = make_regex(r"(\r\n|\n|\r)")
_multiline_whitespace = make_regex(r"\s*", extra_flags=re.MULTILINE)
_whitespace = make_regex(r"[^\S\r\n]*")
_export = make_regex(r"(?:export[^\S\r\n]+)?")
_single_quoted_key = make_regex("'([^']+)'")
_unquoted_key = make_regex(r"([^=\#\s]+)")
_equal_sign = make_regex(r"(=[^\S\r\n]*)")
_single_quoted_value = make_regex(r"'((?:\\'|[^'])*)'")
_double_quoted_value = make_regex(r'"((?:\\"|[^"])*)"')
_unquoted_value = make_regex(r"([^\r\n]*)")
_comment = make_regex(r"(?:[^\S\r\n]*#[^\r\n]*)?")
_end_of_line = make_regex(r"[^\S\r\n]*(?:\r\n|\n|\r|$)")
_rest_of_line = make_regex(r"[^\r\n]*(?:\r|\n|\r\n)?")
_double_quote_escapes = make_regex(r"\\[\\'\"abfnrtv]")
_single_quote_escapes = make_regex(r"\\[\\']")


@dataclasses.dataclass(frozen=True)
class Original:
    """Represents the position of the original string in the file."""

    string: str
    line: int


@dataclasses.dataclass(frozen=True)
class Binding:
    """Represents a key-value pair from a dotenv file."""

    key: str | None
    value: str | None
    original: Original
    error: bool


@dataclasses.dataclass
class Position:
    """Model representing the cursor's position in the file."""

    chars: int
    line: int

    @classmethod
    def start(cls) -> Position:
        """Get the starting position (line 1, character 0)."""

        return cls(chars=0, line=1)

    def set(self, other: Position) -> None:
        """Set the current position to another position."""

        self.chars = other.chars
        self.line = other.line

    def advance(self, string: str) -> None:
        """Advance the position by the length of the given string."""

        self.chars += len(string)
        self.line += len(re.findall(_newline, string))


class Reader:
    """Reader for processing a dotenv file stream."""

    def __init__(self, stream: IO[str]) -> None:
        """Initialize the reader with a stream of data."""

        self.string = stream.read()
        self.position = Position.start()
        self.mark = Position.start()

    def has_next(self) -> bool:
        """Check if there are more characters to read."""

        return self.position.chars < len(self.string)

    def set_mark(self) -> None:
        """Set a mark at the current position."""

        self.mark.set(self.position)

    def get_marked(self) -> Original:
        """Get the portion of the string marked between the current position and the mark."""

        return Original(
            string=self.string[self.mark.chars : self.position.chars],
            line=self.mark.line,
        )

    def peek(self, count: int) -> str:
        """Peek ahead in the string without advancing the cursor."""

        return self.string[self.position.chars : self.position.chars + count]

    def read_regex(self, regex: re.Pattern[str]) -> tuple[str, ...]:
        """Match and read a portion of the string using the given regex."""

        match = regex.match(self.string, self.position.chars)

        if match is None:
            msg = "Pattern not found."
            raise PoetryPluginDotenvPatternError(msg)

        matched = self.string[match.start() : match.end()]
        self.position.advance(matched)
        return match.groups()


def decode_match(match: re.Match[str]) -> str:
    """Decode a match using unicode escapes."""

    return codecs.decode(match.group(0), "unicode-escape")


def decode_escapes(regex: re.Pattern[str], string: str) -> str:
    """Decode escape sequences in the string using the provided regex."""

    return regex.sub(decode_match, string)


def parse_key(reader: Reader) -> str | None:
    """Parse the key part of a dotenv binding."""

    char = reader.peek(1)

    if char == "#":
        return None

    if char == "'":
        (key,) = reader.read_regex(_single_quoted_key)

    else:
        (key,) = reader.read_regex(_unquoted_key)

    return key


def parse_unquoted_value(reader: Reader) -> str:
    """Parse a value that is not quoted."""

    (part,) = reader.read_regex(_unquoted_value)
    return re.sub(r"\s+#.*", "", part).rstrip()


def parse_value(reader: Reader) -> str:
    """Parse a value from a dotenv binding."""

    char = reader.peek(1)

    if char == "'":
        (value,) = reader.read_regex(_single_quoted_value)
        return decode_escapes(_single_quote_escapes, value)

    if char == '"':
        (value,) = reader.read_regex(_double_quoted_value)
        return decode_escapes(_double_quote_escapes, value)

    if char in {"", "\n", "\r"}:
        return ""

    return parse_unquoted_value(reader)


def parse_binding(reader: Reader) -> Binding:
    """Parse a single dotenv binding (key-value pair)."""

    reader.set_mark()

    try:
        reader.read_regex(_multiline_whitespace)

        if not reader.has_next():
            return Binding(
                key=None,
                value=None,
                original=reader.get_marked(),
                error=False,
            )

        reader.read_regex(_export)
        key = parse_key(reader)
        reader.read_regex(_whitespace)

        if reader.peek(1) == "=":
            reader.read_regex(_equal_sign)
            value = parse_value(reader)

        else:
            value = None

        reader.read_regex(_comment)
        reader.read_regex(_end_of_line)

        return Binding(
            key=key,
            value=value,
            original=reader.get_marked(),
            error=False,
        )

    except PoetryPluginDotenvPatternError:
        reader.read_regex(_rest_of_line)
        return Binding(
            key=None,
            value=None,
            original=reader.get_marked(),
            error=True,
        )


def parse_stream(stream: IO[str]) -> Iterator[Binding]:
    """Parse a dotenv stream and yield key-value bindings."""

    reader = Reader(stream)

    while reader.has_next():
        yield parse_binding(reader)



"""Module that contains the core functionality of the plugin."""

from __future__ import annotations

import os

from typing import TYPE_CHECKING

from cleo.events.console_events import COMMAND
from poetry.console.commands.env_command import EnvCommand
from poetry.plugins.application_plugin import ApplicationPlugin

from poetry_plugin_dotenv import dotenv
from poetry_plugin_dotenv.config import Config
from poetry_plugin_dotenv.logger import Logger


if TYPE_CHECKING:  # pragma: no cover
    from cleo.events.console_command_event import ConsoleCommandEvent
    from poetry.console.application import Application


class DotenvPlugin(ApplicationPlugin):
    """Plugin that automatically loads environment variables from a dotenv file.

    This plugin automatically loads environment variables from a dotenv file into the environment
    before the `poetry` commands are executed.
    """

    def activate(self, application: Application) -> None:  # pragma: no cover
        """Activate the plugin."""

        application.event_dispatcher.add_listener(COMMAND, listener=self.load)  # type: ignore[union-attr, arg-type]

    def load(self, event: ConsoleCommandEvent, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Load a dotenv file according to the provided configuration."""

        working_dir = event.io.input.option("directory") or os.path.curdir

        logger = Logger(event)
        config = Config(working_dir)

        if not isinstance(event.command, EnvCommand):
            return  # pragma: nocover

        if config.ignore:
            logger.warning("Not loading environment variables. Ignored by configuration")
            return

        filepath = self._determine_filepath(config, working_dir)

        if filepath == "":
            logger.warning("Not loading environment variables. No valid filepath")
            return

        if os.path.isfile(filepath):
            msg = f"Loading environment variables: <fg=green>{filepath}</>"
            logger.info(msg)
            dotenv.core.load(filepath)

        else:
            msg = f"Could not load environment variables. The file does not exist: {filepath}"
            logger.error(msg)

    def _determine_filepath(self, config: Config, working_dir: str) -> str:
        """Determine the appropriate filepath for the dotenv file."""

        if config.location:
            return (
                os.path.abspath(config.location)
                if os.path.isabs(config.location)
                else os.path.join(working_dir, config.location)
            )

        return dotenv.core.find(usecwd=True)
